package co.dapi.dapi;

import android.app.Activity;
import android.app.Application;
import android.util.Log;

import androidx.annotation.NonNull;
import androidx.annotation.VisibleForTesting;

import com.google.gson.Gson;
import com.google.gson.reflect.TypeToken;

import org.jetbrains.annotations.NotNull;
import org.json.JSONObject;

import java.lang.reflect.Type;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;

import co.dapi.connect.core.base.Dapi;
import co.dapi.connect.core.callbacks.DapiAccountSelectionCallback;
import co.dapi.connect.core.callbacks.DapiAccountSelectionResult;
import co.dapi.connect.core.callbacks.DapiConnectCallback;
import co.dapi.connect.core.callbacks.DapiConnectResult;
import co.dapi.connect.core.callbacks.DapiTransferCallback;
import co.dapi.connect.core.callbacks.DapiTransferResult;
import co.dapi.connect.data.endpoint_models.DapiAccountsResponse;
import co.dapi.connect.data.endpoint_models.DapiCardsResponse;
import co.dapi.connect.data.models.DapiBeneficiary;
import co.dapi.connect.data.models.DapiColor;
import co.dapi.connect.data.models.DapiConfigurations;
import co.dapi.connect.data.models.DapiConnection;
import co.dapi.connect.data.models.DapiEnvironment;
import co.dapi.connect.data.models.DapiError;
import co.dapi.connect.data.models.DapiLanguage;
import co.dapi.connect.data.models.DapiTheme;
import co.dapi.connect.data.models.DapiThemeConfigurations;
import co.dapi.connect.data.models.DapiTransactionsType;
import co.dapi.connect.data.models.DapiWireBeneficiary;
import co.dapi.connect.data.models.LinesAddress;
import io.flutter.embedding.engine.plugins.FlutterPlugin;
import io.flutter.embedding.engine.plugins.activity.ActivityAware;
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding;
import io.flutter.plugin.common.EventChannel;
import io.flutter.plugin.common.MethodCall;
import io.flutter.plugin.common.MethodChannel;
import io.flutter.plugin.common.MethodChannel.MethodCallHandler;
import io.flutter.plugin.common.MethodChannel.Result;
import kotlin.Unit;
import kotlin.jvm.functions.Function1;

/**
 * DapiPlugin
 */
public class DapiPlugin implements FlutterPlugin, MethodCallHandler, ActivityAware {
    /// This local reference serves to register the plugin with the Flutter Engine and unregister it
    /// when the Flutter Engine is detached from the Activity
    private MethodChannel channel;
    @VisibleForTesting
    public Application application;
    private Activity activity;
    private EventChannel.EventSink connectionSuccessfulEvents;
    private EventChannel.EventSink connectionFailureEvents;
    private EventChannel.EventSink bankRequestEvents;
    private EventChannel.EventSink connectDismissedEvents;
    private EventChannel.EventSink transferUiDismissedEvents;
    private EventChannel.EventSink transferUiWillSendEvents;

    private final DapiConnectCallback connectCallback = new DapiConnectCallback() {
        @Override
        public void onConnectionSuccessful(@NonNull DapiConnectResult.Success result) {
            if (activity != null && connectionSuccessfulEvents != null) {
                activity.runOnUiThread(() -> {
                    HashMap<String, Object> data = new HashMap<>();
                    data.put("bankID", result.getConnection().getBankId());
                    data.put("userID", result.getConnection().getUserID());
                    connectionSuccessfulEvents.success(data);
                });
            }
        }

        @Override
        public void onConnectionFailure(@NonNull DapiConnectResult.Error result) {
            if (connectionFailureEvents != null) {
                HashMap<String, Object> data = new HashMap<>();
                data.put("bankID", result.getBankID());
                data.put("error", result.getError().getType() + "&&&" + result.getError().getMessage());
                connectionFailureEvents.success(data);
            }
        }

        @Override
        public void onBankRequest(@NonNull DapiConnectResult.BankRequest result) {
            if (bankRequestEvents != null) {
                HashMap<String, Object> data = new HashMap<>();
                data.put("bankName", result.getBankName());
                data.put("iban", result.getIban());
                bankRequestEvents.success(data);
            }
        }

        @Override
        public void onDismissed() {
            if (connectDismissedEvents != null) {
                connectDismissedEvents.success(null);
            }
        }
    };

    @VisibleForTesting
    public DapiTransferCallback transferCallback;

    @Override
    public void onAttachedToEngine(@NonNull FlutterPluginBinding flutterPluginBinding) {
        channel = new MethodChannel(flutterPluginBinding.getBinaryMessenger(), "dapi");
        channel.setMethodCallHandler(this);

        EventChannel successfulConnectionChannel = new EventChannel(flutterPluginBinding.getBinaryMessenger(), "co.dapi.connect/connect_success_channel");
        EventChannel failureConnectionChannel = new EventChannel(flutterPluginBinding.getBinaryMessenger(), "co.dapi.connect/connect_failure_channel");
        EventChannel bankRequestChannel = new EventChannel(flutterPluginBinding.getBinaryMessenger(), "co.dapi.connect/bank_request_channel");
        EventChannel connectDismissedChannel = new EventChannel(flutterPluginBinding.getBinaryMessenger(), "co.dapi.connect/connect_dismissed_channel");
        EventChannel transferUiDismissedChannel = new EventChannel(flutterPluginBinding.getBinaryMessenger(), "co.dapi.connect/transfer_ui_dismissed_channel");
        EventChannel transferUiWillSendChannel = new EventChannel(flutterPluginBinding.getBinaryMessenger(), "co.dapi.connect/transfer_ui_will_send_channel");

        final SuccessfulConnectionStreamHandler successfulConnectionStreamHandler = new SuccessfulConnectionStreamHandler();
        final FailureConnectionStreamHandler failureConnectionStreamHandler = new FailureConnectionStreamHandler();
        final BankRequestStreamHandler bankRequestStreamHandler = new BankRequestStreamHandler();
        final ConnectDismissedStreamHandler connectDismissedStreamHandler = new ConnectDismissedStreamHandler();
        final TransferUiDismissedStreamHandler transferUiDismissedStreamHandler = new TransferUiDismissedStreamHandler();
        final TransferUiWillSendStreamHandler transferUiWillSendStreamHandler = new TransferUiWillSendStreamHandler();

        successfulConnectionChannel.setStreamHandler(successfulConnectionStreamHandler);
        failureConnectionChannel.setStreamHandler(failureConnectionStreamHandler);
        bankRequestChannel.setStreamHandler(bankRequestStreamHandler);
        connectDismissedChannel.setStreamHandler(connectDismissedStreamHandler);
        transferUiDismissedChannel.setStreamHandler(transferUiDismissedStreamHandler);
        transferUiWillSendChannel.setStreamHandler(transferUiWillSendStreamHandler);
    }

    @Override
    public void onAttachedToActivity(@NonNull @NotNull ActivityPluginBinding binding) {
        activity = binding.getActivity();
        application = activity.getApplication();
    }

    @Override
    public void onDetachedFromActivityForConfigChanges() {
        application = null;
        activity = null;
    }

    @Override
    public void onReattachedToActivityForConfigChanges(@NonNull @NotNull ActivityPluginBinding binding) {
        activity = binding.getActivity();
        application = activity.getApplication();
    }

    @Override
    public void onDetachedFromActivity() {
        application = null;
        activity = null;
    }

    @SuppressWarnings("unchecked")
    @Override
    public void onMethodCall(@NonNull MethodCall call, @NonNull Result result) {
        List<Object> args = (List<Object>) call.arguments;
        switch (call.method) {
            case "startWithAppKey:clientUserID:configurations:":
                Map<String, Object> configurations = null;
                if (args.size() > 2 && args.get(2) != null) {
                    configurations = (Map<String, Object>) args.get(2);
                }
                start((String) args.get(0), (String) args.get(1), configurations, result);
                break;
            case "isStarted":
                isStarted(result);
                break;
            case "setConfigurations:":
                setConfigurations((Map<String, Object>) args.get(0));
                break;
            case "configurations":
                configurations(result);
                break;
            case "presentConnect:":
                String bankID = null;
                if (args.size() > 0) {
                    bankID = (String) args.get(0);
                }
                presentConnect(bankID);
                break;
            case "dismissConnect":
                dismissConnect();
                break;
            case "dismissTransfer":
                dismissTransfer();
                break;
            case "dismissMFA":
                dismissMFA();
                break;
            case "setClientUserID:":
                setClientUserID((String) args.get(0));
                break;
            case "clientUserID":
                clientUserID(result);
                break;
            case "getConnections":
                getConnections(result);
                break;
            case "createConnection:":
                createConnection((String) args.get(0), result);
                break;
            case "getParameters:":
                getParameters((String) args.get(0), result);
                break;
            case "getIdentity:":
                getIdentity((String) args.get(0), result);
                break;
            case "getAccounts:":
                getAccounts((String) args.get(0), result);
                break;
            case "getCards:":
                getCards((String) args.get(0), result);
                break;
            case "getTransactionsForAccount:accountID:startDate:endDate:type:":
                getTransactionsForAccount((String) args.get(0), (String) args.get(1), (Long) args.get(2), (Long) args.get(3), (String) args.get(4), result);
                break;
            case "getTransactionsForCard:cardID:startDate:endDate:type:":
                getTransactionsForCard((String) args.get(0), (String) args.get(1), (Long) args.get(2), (Long) args.get(3), (String) args.get(4), result);
                break;
            case "getAccountsMetadata:":
                getAccountsMetaData((String) args.get(0), result);
                break;
            case "delete:":
                delete((String) args.get(0), result);
                break;
            case "getBeneficiaries:":
                getBeneficiaries((String) args.get(0), result);
                break;
            case "createBeneficiary:beneficiary:":
                createBeneficiary((String) args.get(0), (HashMap<String, Object>) args.get(1), result);
                break;
            case "createTransfer:accountID:beneficiary:amount:remark:":
                String accountID = null;
                HashMap<String, Object> beneficiary = null;
                double amount = 0.0;
                String remark = null;
                if (args.size() > 1 && args.get(1) != null) {
                    accountID = (String) args.get(1);
                }
                if (args.size() > 2 && args.get(2) != null) {
                    beneficiary = (HashMap<String, Object>) args.get(2);
                }
                if (args.size() > 3 && args.get(3) != null) {
                    amount = (double) args.get(3);
                }
                if (args.size() > 4 && args.get(4) != null) {
                    remark = (String) args.get(4);
                }
                createTransfer((String) args.get(0), accountID, beneficiary, amount, remark, result);
                break;
            case "createTransferToExistingBeneficiary:accountID:receiverID:amount:remark:":
                createTransferToExistingBeneficiary((String) args.get(0), (String) args.get(1), (String) args.get(2), (double) args.get(3), (String) args.get(4), result);
                break;
            case "getWireBeneficiaries:":
                getWireBeneficiaries((String) args.get(0), result);
                break;
            case "createWireBeneficiary:beneficiary:":
                createWireBeneficiary((String) args.get(0), (HashMap<String, Object>) args.get(1), result);
                break;
            case "createWireTransfer:accountID:beneficiary:amount:remark:":
                String wireAccountId = null;
                HashMap<String, Object> wireBeneficiary = null;
                double wireAmount = 0.0;
                String wireRemark = null;
                if (args.size() > 1 && args.get(1) != null) {
                    wireAccountId = (String) args.get(1);
                }
                if (args.size() > 2 && args.get(2) != null) {
                    wireBeneficiary = (HashMap<String, Object>) args.get(2);
                }
                if (args.size() > 3 && args.get(3) != null) {
                    wireAmount = (double) args.get(3);
                }
                if (args.size() > 4 && args.get(4) != null) {
                    wireRemark = (String) args.get(4);
                }
                createWireTransfer((String) args.get(0), wireAccountId, wireBeneficiary, wireAmount, wireRemark, result);
                break;
            case "createWireTransferToExistingBeneficiary:accountID:receiverID:amount:remark:":
                createWireTransferToExistingBeneficiary((String) args.get(0), (String) args.get(1), (String) args.get(2), (double) args.get(3), (String) args.get(4), result);
                break;
            case "createACHPullTransfer:description:accountID:amount:":
                createACHPullTransfer((String) args.get(0), (String) args.get(1), (String) args.get(2), (double) args.get(3), result);
                break;
            case "presentAccountSelection:":
                presentAccountSelection((String) args.get(0), result);
                break;
            default:
                result.notImplemented();
                break;
        }
    }

    @Override
    public void onDetachedFromEngine(@NonNull FlutterPluginBinding binding) {
        channel.setMethodCallHandler(null);
    }

    @VisibleForTesting
    public void start(String appKey, String clientUserID, Map<String, Object> configurationsMap, Result result) {
        DapiConfigurations configurations = getConfigurationsFromMap(configurationsMap);
        Dapi.start(application, appKey, clientUserID, configurations, () -> {
            result.success(null);
            return Unit.INSTANCE;
        }, (error) -> {
            reject(result, "-100", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void setConfigurations(Map<String, Object> configurationsMap) {
        Dapi.setConfigurations(getConfigurationsFromMap(configurationsMap));
    }

    @VisibleForTesting
    public void configurations(Result result) {
        result.success(getMap(Dapi.getConfigurations()));
    }

    @VisibleForTesting
    public void presentConnect(String bankID) {
        Dapi.presentConnect(bankID);
    }

    @VisibleForTesting
    public void isStarted(Result result) {
        result.success(Dapi.isStarted());
    }

    @VisibleForTesting
    public void dismissConnect() {
        Dapi.dismissConnect();
    }

    @VisibleForTesting
    public void dismissTransfer() {
        Dapi.dismissTransfer();
    }

    @VisibleForTesting
    public void dismissMFA() {
        Dapi.dismissMFA();
    }

    @VisibleForTesting
    public void setClientUserID(String clientUserID) {
        Dapi.setClientUserID(clientUserID);
    }

    @VisibleForTesting
    public void clientUserID(Result result) {
        result.success(Dapi.getClientUserID());
    }

    @VisibleForTesting
    public void getConnections(Result result) {
        Dapi.getConnections(connections -> {
            ArrayList<HashMap<String, Object>> connectionsArray = getBridgeConnections(connections);
            result.success(connectionsArray);
            return Unit.INSTANCE;
        }, error -> {
            reject(result, "-200", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void createConnection(String jsonConnectionParams, Result result) {
        DapiConnection.create(jsonConnectionParams, connection -> {
            result.success(getBridgeConnection(connection));
            return Unit.INSTANCE;
        }, error -> {
            reject(result, "-201", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void getParameters(String userID, Result result) {
        getOperatingConnection(userID, connection -> {
            connection.getParameters(params -> {
                result.success(params.toString());
                return Unit.INSTANCE;
            }, error -> {
                reject(result, "-202", error);
                return Unit.INSTANCE;
            });
            return Unit.INSTANCE;
        }, error -> {
            reject(result, "-202", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void presentAccountSelection(String userID, Result result) {
        getOperatingConnection(userID, connection -> {
            connection.presentAccountSelection(new DapiAccountSelectionCallback() {
                @Override
                public void onSelected(@NonNull DapiAccountSelectionResult.Success successResult) {
                    result.success(successResult.getAccount().getId());
                }

                @Override
                public void onFailure(@NonNull DapiAccountSelectionResult.Error errorResult) {
                    reject(result, "-800", errorResult.getError());
                }

                @Override
                public void onDismissed() {
                    DapiError error = new DapiError(DapiError.USER_CANCELLED, "Account selection was cancelled", "Account selection was cancelled", null, null, null);
                    reject(result, "-800", error);
                }
            });
            return Unit.INSTANCE;
        }, error -> {
            reject(result, "-800", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void getIdentity(String userID, Result result) {
        getOperatingConnection(userID, connection -> {
            connection.getIdentity(identity -> {
                result.success(getMap(identity));
                return Unit.INSTANCE;
            }, error -> {
                reject(result, "-300", error);
                return Unit.INSTANCE;
            });
            return null;
        }, error -> {
            reject(result, "-300", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void getAccounts(String userID, Result result) {
        getOperatingConnection(userID, connection -> {
            connection.getAccounts(accounts -> {
                result.success(getMap(accounts));
                return Unit.INSTANCE;
            }, error -> {
                reject(result, "-301", error);
                return Unit.INSTANCE;
            });
            return Unit.INSTANCE;
        }, error -> {
            reject(result, "-301", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void getCards(String userID, Result result) {
        getOperatingConnection(userID, connection -> {
            connection.getCards(cards -> {
                result.success(getMap(cards));
                return Unit.INSTANCE;
            }, error -> {
                reject(result, "-302", error);
                return Unit.INSTANCE;
            });
            return null;
        }, error -> {
            reject(result, "-302", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void getTransactionsForAccount(
            String userID,
            String accountID,
            long startDate,
            long endDate,
            String type,
            Result result
    ) {
        Date startDateObject = new Date(startDate);
        Date endDateObject = new Date(endDate);
        DapiTransactionsType nativeType;
        if (type.toLowerCase(Locale.ROOT).equals("enriched")) {
            nativeType = DapiTransactionsType.ENRICHED;
        } else if (type.toLowerCase(Locale.ROOT).equals("categorized")) {
            nativeType = DapiTransactionsType.CATEGORIZED;
        } else {
            nativeType = DapiTransactionsType.DEFAULT;
        }
        getOperatingConnection(userID, connection -> {
            DapiAccountsResponse.DapiAccount account = getDapiAccount(accountID, connection);
            connection.getTransactions(account, startDateObject, endDateObject, nativeType, transactions -> {
                result.success(getMap(transactions));
                return Unit.INSTANCE;
            }, error -> {
                reject(result, "-303", error);
                return Unit.INSTANCE;
            });
            return null;
        }, error -> {
            reject(result, "-303", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void getTransactionsForCard(
            String userID,
            String cardID,
            long startDate,
            long endDate,
            String type,
            Result result
    ) {
        Date startDateObject = new Date(startDate);
        Date endDateObject = new Date(endDate);
        DapiTransactionsType nativeType;
        if (type.toLowerCase(Locale.ROOT).equals("enriched")) {
            nativeType = DapiTransactionsType.ENRICHED;
        } else if (type.toLowerCase(Locale.ROOT).equals("categorized")) {
            nativeType = DapiTransactionsType.CATEGORIZED;
        } else {
            nativeType = DapiTransactionsType.DEFAULT;
        }
        getOperatingConnection(userID, connection -> {
            DapiCardsResponse.DapiCard card = getDapiCard(cardID, connection);
            connection.getTransactions(card, startDateObject, endDateObject, nativeType, transactions -> {
                result.success(getMap(transactions));
                return Unit.INSTANCE;
            }, error -> {
                reject(result, "-304", error);
                return Unit.INSTANCE;
            });
            return Unit.INSTANCE;
        }, error -> {
            reject(result, "-304", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void getAccountsMetaData(String userID, Result result) {
        getOperatingConnection(userID, connection -> {
            connection.getAccountsMetaData(metadata -> {
                result.success(getMap(metadata));
                return Unit.INSTANCE;
            }, error -> {
                reject(result, "-400", error);
                return Unit.INSTANCE;
            });
            return Unit.INSTANCE;
        }, error -> {
            reject(result, "-400", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void delete(
            String userID,
            Result result
    ) {
        getOperatingConnection(userID, connection -> {
            connection.delete(delink -> {
                result.success(getMap(delink));
                return Unit.INSTANCE;
            }, error -> {
                reject(result, "-500", error);
                return Unit.INSTANCE;
            });
            return null;
        }, error -> {
            reject(result, "-500", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void getBeneficiaries(String userID, Result result) {
        getOperatingConnection(userID, connection -> {
            connection.getBeneficiaries(beneficiaries -> {
                result.success(getMap(beneficiaries));
                return Unit.INSTANCE;
            }, error -> {
                reject(result, "-600", error);
                return Unit.INSTANCE;
            });
            return Unit.INSTANCE;
        }, error -> {
            reject(result, "-600", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void createBeneficiary(
            String userID,
            HashMap<String, Object> beneficiaryMap
            , Result result) {
        getOperatingConnection(userID, connection -> {
            connection.createBeneficiary(getBeneficiary(beneficiaryMap), createBeneficiary -> {
                result.success(getMap(createBeneficiary));
                return Unit.INSTANCE;
            }, error -> {
                reject(result, "-601", error);
                return Unit.INSTANCE;
            });
            return Unit.INSTANCE;
        }, error -> {
            reject(result, "-601", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void createTransfer(
            String userID,
            String accountID,
            HashMap<String, Object> beneficiaryMap,
            double amount,
            String remark,
            Result result
    ) {
        setTransferCallback(result);
        getOperatingConnection(userID, connection -> {
            DapiAccountsResponse.DapiAccount account = getDapiAccount(accountID, connection);
            DapiBeneficiary beneficiary = getBeneficiary(beneficiaryMap);
            connection.createTransfer(account, beneficiary, amount, remark);
            return Unit.INSTANCE;
        }, error -> {
            reject(result, "-602", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void createTransferToExistingBeneficiary(
            String userID,
            String accountID,
            String receiverID,
            double amount,
            String remark,
            Result result
    ) {
        setTransferCallback(result);
        getOperatingConnection(userID, connection -> {
            DapiAccountsResponse.DapiAccount account = getDapiAccount(accountID, connection);
            connection.createTransferToExistingBeneficiary(account, receiverID, amount, remark);
            return Unit.INSTANCE;
        }, error -> {
            reject(result, "-603", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void getWireBeneficiaries(String userID, Result result) {
        getOperatingConnection(userID, connection -> {
            connection.getWireBeneficiaries(beneficiaries -> {
                result.success(getMap(beneficiaries));
                return Unit.INSTANCE;
            }, error -> {
                reject(result, "-700", error);
                return Unit.INSTANCE;
            });
            return Unit.INSTANCE;
        }, error -> {
            reject(result, "-700", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void createWireBeneficiary(
            String userID,
            HashMap<String, Object> beneficiaryMap
            , Result result) {
        getOperatingConnection(userID, connection -> {
            connection.createWireBeneficiary(getWireBeneficiary(beneficiaryMap), createBeneficiary -> {
                result.success(getMap(createBeneficiary));
                return Unit.INSTANCE;
            }, error -> {
                reject(result, "-701", error);
                return Unit.INSTANCE;
            });
            return Unit.INSTANCE;
        }, error -> {
            reject(result, "-701", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void createWireTransfer(
            String userID,
            String accountID,
            HashMap<String, Object> beneficiaryMap,
            double amount,
            String remark,
            Result result
    ) {
        setTransferCallback(result);
        getOperatingConnection(userID, connection -> {
            DapiAccountsResponse.DapiAccount account = getDapiAccount(accountID, connection);
            DapiWireBeneficiary beneficiary = getWireBeneficiary(beneficiaryMap);
            connection.createWireTransfer(beneficiary, account, amount, remark);
            return Unit.INSTANCE;
        }, error -> {
            reject(result, "-702", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void createWireTransferToExistingBeneficiary(
            String userID,
            String accountID,
            String receiverID,
            double amount,
            String remark,
            Result result
    ) {
        setTransferCallback(result);
        getOperatingConnection(userID, connection -> {
            DapiAccountsResponse.DapiAccount account = getDapiAccount(accountID, connection);
            connection.createWireTransferToExistingBeneficiary(account, receiverID, amount, remark);
            return Unit.INSTANCE;
        }, error -> {
            reject(result, "-703", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    public void createACHPullTransfer(
            String userID,
            String description,
            String accountID,
            double amount,
            Result result
    ) {
        setTransferCallback(result);
        getOperatingConnection(userID, connection -> {
            DapiAccountsResponse.DapiAccount account = getDapiAccount(accountID, connection);
            connection.createACHPullTransfer(description, account, amount);
            return Unit.INSTANCE;
        }, error -> {
            reject(result, "-702", error);
            return Unit.INSTANCE;
        });
    }

    @VisibleForTesting
    void reject(Result result, String errorCode, DapiError error) {
        JSONObject errorResult = new JSONObject();
        try {
            errorResult.put("message", error.getMessage());
            errorResult.put("type", error.getType());
            errorResult.put("operationID", error.getOperationID());
            if (error.getCoolDownPeriod() != null) {
                errorResult.put("unit", error.getCoolDownPeriod().getUnit());
                errorResult.put("value", error.getCoolDownPeriod().getValue());
            }
            result.error(errorCode, errorResult.toString(), getMap(error));
        } catch (Exception e) {
            e.printStackTrace();
            HashMap<String, String> map = new HashMap<>();
            map.put("message", e.getLocalizedMessage());
            result.error(errorCode, new JSONObject(map).toString(), getMap(error));
        }
    }

    @SuppressWarnings({"ConstantConditions", "unchecked"})
    private DapiConfigurations getConfigurationsFromMap(Map<String, Object> configurations) {
        if (configurations == null) {
            return null;
        }
        HashMap<String, Object> extraQueryParameters;
        HashMap<String, Object> extraHeaderFields;
        HashMap<String, Object> extraBody;
        DapiEnvironment environment;
        String[] countries;
        boolean showLogos = true;
        boolean showExperimentalBanks = false;
        boolean showCloseButton = true;
        boolean showAddButton = false;
        boolean showTransferSuccessfulResult = true;
        boolean showTransferErrorResult = true;
        String postSuccessfulConnectionLoadingText = "Verifying your account..";
        DapiThemeConfigurations theme;
        DapiLanguage language;

        if (configurations.get("countries") != null) {
            ArrayList<String> countriesList = (ArrayList<String>) configurations.get("countries");
            countries = new String[countriesList.size()];
            for (int i = 0; i < countriesList.size(); i++)
                countries[i] = countriesList.get(i);
        } else {
            countries = new String[]{};
        }

        if (configurations.get("showLogos") != null) {
            showLogos = (boolean) configurations.get("showLogos");
        }

        if (configurations.get("showExperimentalBanks") != null) {
            showExperimentalBanks = (boolean) configurations.get("showExperimentalBanks");
        }

        if (configurations.get("showCloseButton") != null) {
            showCloseButton = (boolean) configurations.get("showCloseButton");
        }

        if (configurations.get("showAddButton") != null) {
            showAddButton = (boolean) configurations.get("showAddButton");
        }

        if (configurations.get("showTransferSuccessfulResult") != null) {
            showTransferSuccessfulResult = (boolean) configurations.get("showTransferSuccessfulResult");
        }

        if (configurations.get("showTransferErrorResult") != null) {
            showTransferErrorResult = (boolean) configurations.get("showTransferErrorResult");
        }

        if (configurations.get("postSuccessfulConnectionLoadingText") != null) {
            postSuccessfulConnectionLoadingText = (String) configurations.get("postSuccessfulConnectionLoadingText");
        }

        if (configurations.get("extraQueryParameters") != null) {
            extraQueryParameters = (HashMap<String, Object>) configurations.get("extraQueryParameters");
        } else {
            extraQueryParameters = new HashMap<>();
        }

        if (configurations.get("extraHeaderFields") != null) {
            extraHeaderFields = (HashMap<String, Object>) configurations.get("extraHeaderFields");
        } else {
            extraHeaderFields = new HashMap<>();
        }

        if (configurations.get("extraBody") != null) {
            extraBody = (HashMap<String, Object>) configurations.get("extraBody");
        } else {
            extraBody = new HashMap<>();
        }

        if (configurations.get("theme") != null) {
            theme = getTheme(configurations);
        } else {
            theme = new DapiThemeConfigurations();
        }

        if (configurations.get("language") != null) {
            String languageString = (String) configurations.get("language");
            if (languageString.toLowerCase(Locale.ROOT).equals("ar")) {
                language = DapiLanguage.AR;
            } else {
                language = DapiLanguage.EN;
            }
        } else {
            language = DapiLanguage.EN;
        }

        if (configurations.get("environment") != null) {
            String environmentString = (String) configurations.get("environment");
            if (environmentString.toLowerCase(Locale.ROOT).equals("sandbox")) {
                environment = DapiEnvironment.SANDBOX;
            } else {
                environment = DapiEnvironment.PRODUCTION;
            }
        } else {
            environment = DapiEnvironment.PRODUCTION;
        }

        return new DapiConfigurations(
                extraBody,
                extraQueryParameters,
                extraHeaderFields,
                environment,
                countries,
                showLogos,
                showExperimentalBanks,
                showCloseButton,
                showAddButton,
                postSuccessfulConnectionLoadingText,
                showTransferSuccessfulResult,
                showTransferErrorResult,
                theme,
                language
        );
    }

    @SuppressWarnings("unchecked")
    @NonNull
    private DapiThemeConfigurations getTheme(Map<String, Object> configurations) {
        DapiThemeConfigurations theme;
        if (configurations.get("theme") == null) {
            return new DapiThemeConfigurations();
        }

        Map<String, Object> themeMap = (Map<String, Object>) configurations.get("theme");
        if (themeMap == null) {
            return new DapiThemeConfigurations();
        }
        Map<String, Object> primaryColor = (Map<String, Object>) themeMap.get("primaryColor");
        String rawEnforceTheme = (String) themeMap.get("enforceTheme");
        String lightMode = null;
        String darkMode = null;
        if (primaryColor != null) {
            lightMode = (String) primaryColor.get("lightMode");
            darkMode = (String) primaryColor.get("darkMode");
        }
        DapiTheme enforceTheme;
        if (rawEnforceTheme != null && rawEnforceTheme.equals("light")) {
            enforceTheme = DapiTheme.LIGHT;
        } else if (rawEnforceTheme != null && rawEnforceTheme.equals("dark")) {
            enforceTheme = DapiTheme.DARK;
        } else {
            enforceTheme = DapiTheme.DYNAMIC;
        }
        theme = new DapiThemeConfigurations(
                enforceTheme,
                new DapiColor(lightMode, darkMode)
        );
        return theme;
    }

    @SuppressWarnings({"unchecked", "ConstantConditions"})
    private DapiBeneficiary getBeneficiary(HashMap<String, Object> beneficiaryMap) {
        if (beneficiaryMap == null) {
            return null;
        }
        HashMap<String, Object> linesMap = (HashMap<String, Object>) beneficiaryMap.get("address");
        String line1 = (String) linesMap.get("line1");
        String line2 = (String) linesMap.get("line2");
        String line3 = (String) linesMap.get("line3");
        LinesAddress linesAddress = new LinesAddress(
                line1,
                line2,
                line3
        );
        String accountNumber = (String) beneficiaryMap.get("accountNumber");
        String name = (String) beneficiaryMap.get("name");
        String bankName = (String) beneficiaryMap.get("bankName");
        String swiftCode = (String) beneficiaryMap.get("swiftCode");
        String iban = (String) beneficiaryMap.get("iban");
        String phoneNumber = (String) beneficiaryMap.get("phoneNumber");
        String country = (String) beneficiaryMap.get("country");
        String branchAddress = (String) beneficiaryMap.get("branchAddress");
        String branchName = (String) beneficiaryMap.get("branchName");
        String nickname = (String) beneficiaryMap.get("nickname");

        return new DapiBeneficiary(
                linesAddress,
                accountNumber,
                name,
                bankName,
                swiftCode,
                iban,
                country,
                branchAddress,
                branchName,
                phoneNumber,
                nickname
        );
    }

    @SuppressWarnings({"unchecked", "ConstantConditions"})
    private DapiWireBeneficiary getWireBeneficiary(HashMap<String, Object> beneficiaryMap) {
        if (beneficiaryMap == null) {
            return null;
        }

        HashMap<String, Object> linesMap = (HashMap<String, Object>) beneficiaryMap.get("address");
        String line1 = (String) linesMap.get("line1");
        String line2 = (String) linesMap.get("line2");
        String line3 = (String) linesMap.get("line3");
        LinesAddress linesAddress = new LinesAddress(
                line1,
                line2,
                line3
        );

        String name = (String) beneficiaryMap.get("name");
        String firstName = (String) beneficiaryMap.get("firstName");
        String lastName = (String) beneficiaryMap.get("lastName");
        String nickname = (String) beneficiaryMap.get("nickname");
        String city = (String) beneficiaryMap.get("city");
        String state = (String) beneficiaryMap.get("state");
        String country = (String) beneficiaryMap.get("country");
        String zipCode = (String) beneficiaryMap.get("zipCode");
        String receiverType = (String) beneficiaryMap.get("receiverType");
        String receiverAccountType = (String) beneficiaryMap.get("receiverAccountType");
        String routingNumber = (String) beneficiaryMap.get("routingNumber");
        String accountNumber = (String) beneficiaryMap.get("accountNumber");


        return new DapiWireBeneficiary(
                name,
                firstName,
                lastName,
                nickname,
                linesAddress,
                city,
                state,
                country,
                zipCode,
                receiverType,
                receiverAccountType,
                routingNumber,
                accountNumber
        );
    }

    @SuppressWarnings("ConstantConditions")
    private DapiAccountsResponse.DapiAccount getDapiAccount(String accountID, DapiConnection connection) {
        for (DapiAccountsResponse.DapiAccount account : connection.getAccounts()) {
            if (account.getId().equals(accountID)) {
                return account;
            }
        }
        return null;
    }

    @SuppressWarnings("ConstantConditions")
    private DapiCardsResponse.DapiCard getDapiCard(String cardID, DapiConnection connection) {
        for (DapiCardsResponse.DapiCard card : connection.getCards()) {
            if (card.getId().equals(cardID)) {
                return card;
            }
        }
        return null;
    }

    private void getOperatingConnection(String userID, Function1<? super DapiConnection, Unit> onSuccess, Function1<? super DapiError, Unit> onFailure) {
        Dapi.getConnections(connections -> {
            for (DapiConnection connection : connections) {
                if (connection.getUserID().equals(userID)) {
                    onSuccess.invoke(connection);
                    break;
                }
            }
            return Unit.INSTANCE;
        }, error -> {
            onFailure.invoke(error);
            return Unit.INSTANCE;
        });
    }

    private void setTransferCallback(Result result) {
        transferCallback = new DapiTransferCallback() {
            @Override
            public void onTransferSuccess(@NonNull DapiTransferResult.Success successResult) {
                HashMap<String, Object> data = new HashMap<>();
                data.put("account", successResult.getAccount().getId());
                data.put("amount", successResult.getAmount());
                data.put("operationID", successResult.getOperationID());
                data.put("remark", successResult.getRemark());
                data.put("reference", successResult.getReference());
                result.success(data);
            }

            @Override
            public void onTransferFailure(@NonNull DapiTransferResult.Error errorResult) {
                reject(result, "-602", errorResult.getError());
            }

            @Override
            public void willTransferAmount(@NonNull DapiTransferResult.PreTransfer preTransferResult) {
                if (transferUiWillSendEvents != null) {
                    HashMap<String, Object> data = new HashMap<>();
                    data.put("account", preTransferResult.getAccount().getId());
                    data.put("amount", preTransferResult.getAmount());
                    transferUiWillSendEvents.success(data);
                }
            }

            @Override
            public void onUiDismissed() {
                if (transferUiDismissedEvents != null) {
                    transferUiDismissedEvents.success(null);
                }
            }
        };
        Dapi.setTransferCallback(transferCallback);
    }

    @NonNull
    private ArrayList<HashMap<String, Object>> getBridgeConnections(List<DapiConnection> connections) {
        ArrayList<HashMap<String, Object>> connectionsArray = new ArrayList<>();
        for (DapiConnection connection : connections) {
            HashMap<String, Object> connectionMap = getBridgeConnection(connection);
            connectionsArray.add(connectionMap);
        }
        return connectionsArray;
    }

    @NonNull
    private HashMap<String, Object> getBridgeConnection(DapiConnection connection) {
        HashMap<String, Object> connectionMap = new HashMap<>();
        connectionMap.put("userID", connection.getUserID());
        connectionMap.put("clientUserID", connection.getClientUserID());
        connectionMap.put("bankID", connection.getBankId());
        connectionMap.put("swiftCode", connection.getSwiftCode());
        connectionMap.put("bankShortName", connection.getName());
        connectionMap.put("bankFullName", connection.getFullName());
        connectionMap.put("country", connection.getCountry());
        connectionMap.put("fullLogo", connection.getFullLogoPng());
        connectionMap.put("halfLogo", connection.getHalfLogoPng());
        connectionMap.put("miniLogo", connection.getMiniLogoPng());
        return connectionMap;
    }

    private Map<String, Object> getMap(Object object) {
        Gson gson = new Gson();
        String json = gson.toJson(object);
        Type type = new TypeToken<Map<String, Object>>() {
        }.getType();
        return gson.fromJson(json, type);
    }

    class SuccessfulConnectionStreamHandler implements EventChannel.StreamHandler {
        @Override
        public void onListen(Object arguments, EventChannel.EventSink events) {
            DapiPlugin.this.connectionSuccessfulEvents = events;
            Dapi.setConnectCallback(DapiPlugin.this.connectCallback);
        }

        @Override
        public void onCancel(Object arguments) {
            Dapi.setConnectCallback(null);
        }
    }

    class FailureConnectionStreamHandler implements EventChannel.StreamHandler {
        @Override
        public void onListen(Object arguments, EventChannel.EventSink events) {
            DapiPlugin.this.connectionFailureEvents = events;
            Dapi.setConnectCallback(DapiPlugin.this.connectCallback);
        }

        @Override
        public void onCancel(Object arguments) {
            Dapi.setConnectCallback(null);
        }
    }

    class BankRequestStreamHandler implements EventChannel.StreamHandler {
        @Override
        public void onListen(Object arguments, EventChannel.EventSink events) {
            DapiPlugin.this.bankRequestEvents = events;
            Dapi.setConnectCallback(DapiPlugin.this.connectCallback);
        }

        @Override
        public void onCancel(Object arguments) {
            Dapi.setConnectCallback(null);
        }
    }

    class ConnectDismissedStreamHandler implements EventChannel.StreamHandler {
        @Override
        public void onListen(Object arguments, EventChannel.EventSink events) {
            DapiPlugin.this.connectDismissedEvents = events;
            Dapi.setConnectCallback(DapiPlugin.this.connectCallback);
        }

        @Override
        public void onCancel(Object arguments) {
            Dapi.setConnectCallback(null);
        }
    }

    class TransferUiDismissedStreamHandler implements EventChannel.StreamHandler {
        @Override
        public void onListen(Object arguments, EventChannel.EventSink events) {
            DapiPlugin.this.transferUiDismissedEvents = events;
        }

        @Override
        public void onCancel(Object arguments) {

        }
    }

    class TransferUiWillSendStreamHandler implements EventChannel.StreamHandler {
        @Override
        public void onListen(Object arguments, EventChannel.EventSink events) {
            DapiPlugin.this.transferUiWillSendEvents = events;
        }

        @Override
        public void onCancel(Object arguments) {

        }
    }

}
