#import "DapiPlugin.h"
#import "DPCCrossSDK.h"

#define NULL_TO_NIL(obj) ({ __typeof__ (obj) __obj = (obj); [__obj isKindOfClass:[NSNull class]] ? nil : obj; })

@interface DapiPlugin () <DAPIConnectDelegate,DAPIAutoFlowDelegate>

@property (nonatomic, strong) NSMutableDictionary<NSString *, FlutterResult> *selectorResult;
@property (nonatomic, assign) BOOL isDapiStarted;

@property (nonatomic, strong) DAPIAutoFlowVC *autoFlow;
@property (nonatomic, strong) DAPIConnectVC *connectVC;
@property (nonatomic, strong) DAPIBankAccountsVC *bankAccountsVC;
@property (nonatomic, strong) NSMutableDictionary *cachedAccounts;

@end

@implementation DapiPlugin
+ (void)registerWithRegistrar:(NSObject<FlutterPluginRegistrar>*)registrar {
    FlutterMethodChannel* channel = [FlutterMethodChannel
                                     methodChannelWithName:@"dapi"
                                     binaryMessenger:[registrar messenger]];
    DapiPlugin* instance = [[DapiPlugin alloc] init];
    [registrar addMethodCallDelegate:instance channel:channel];
    
    FlutterEventChannel *successfulConnectionEventChannel = [FlutterEventChannel eventChannelWithName:@"co.dapi.connect/connect_success_channel" binaryMessenger:[registrar messenger]];
    instance.successfulConnectionStream = [[SuccessfulConnectionStream alloc] init];
    [successfulConnectionEventChannel setStreamHandler:instance.successfulConnectionStream];
    
    FlutterEventChannel *failureConnectionEventChannel = [FlutterEventChannel eventChannelWithName:@"co.dapi.connect/connect_failure_channel" binaryMessenger:[registrar messenger]];
    instance.failureConnectionStream = [[FailureConnectionStream alloc] init];
    [failureConnectionEventChannel setStreamHandler:instance.failureConnectionStream];
    
    FlutterEventChannel *bankRequestEventChannel = [FlutterEventChannel eventChannelWithName:@"co.dapi.connect/bank_request_channel" binaryMessenger:[registrar messenger]];
    instance.bankRequestStream = [[BankRequestStream alloc] init];
    [bankRequestEventChannel setStreamHandler:instance.bankRequestStream];
    
    FlutterEventChannel *connectDismissedEventChannel = [FlutterEventChannel eventChannelWithName:@"co.dapi.connect/connect_dismissed_channel" binaryMessenger:[registrar messenger]];
    instance.connectDismissedStream = [[ConnectDismissedStream alloc] init];
    [connectDismissedEventChannel setStreamHandler:instance.connectDismissedStream];
    
    FlutterEventChannel *transferUiDismissedEventChannel = [FlutterEventChannel eventChannelWithName:@"co.dapi.connect/transfer_ui_dismissed_channel" binaryMessenger:[registrar messenger]];
    instance.transferUiDismissedStream = [[TransferUiDismissedStream alloc] init];
    [transferUiDismissedEventChannel setStreamHandler:instance.transferUiDismissedStream];
    
    FlutterEventChannel *transferUiWillSendEventChannel = [FlutterEventChannel eventChannelWithName:@"co.dapi.connect/transfer_ui_will_send_channel" binaryMessenger:[registrar messenger]];
    instance.transferUiWillSendStream = [[TransferUiWillSendStream alloc] init];
    [transferUiWillSendEventChannel setStreamHandler:instance.transferUiWillSendStream];
}

- (instancetype)init
{
    self = [super init];
    if (self) {
        self.selectorResult = [NSMutableDictionary dictionary];
    }
    return self;
}

- (void)dapiConnect:(DAPIConnectVC * _Nonnull)dapiConnect
didFailConnectingWith:(NSString * _Nonnull)error {
    NSDictionary *body = @{
        //        @"bankID": bankID ?: [NSNull null],
        @"error": error ?: [NSNull null]
    };
    [self.failureConnectionStream send: body];

}

- (void)dapiConnect:(DAPIConnectVC * _Nonnull)dapiConnect
didSuccessfullyConnectTo:(NSString * _Nullable)bankID
     bankConnection:(DAPIBankConnection * _Nonnull)bankConnection {

    NSDictionary * dic = @{
        @"bankID": bankID,
        @"userID": bankConnection.userID
    };
    [self.successfulConnectionStream send: dic];

}

- (void)dapiConnectUserDidCancel:(DAPIConnectVC * _Nonnull)dapiConnect {
    [self.connectDismissedStream send: nil];
}



- (void)handleMethodCall:(FlutterMethodCall*)call result:(FlutterResult)result {
    BOOL isImplemented = NO;
    SEL selector = NSSelectorFromString(call.method);
    if ([self respondsToSelector:selector]) {
        isImplemented = YES;
        
        [self.selectorResult setObject:result forKey:call.method];
        
        NSInvocation *invocation = [NSInvocation invocationWithMethodSignature:[self methodSignatureForSelector:selector]];
        invocation.selector = selector;
        invocation.target = self;
        
        /*
         * Indices 0 and 1 indicate the hidden arguments self and _cmd,
         * respectively; you should set these values directly with the target and selector properties.
         * Use indices 2 and greater for the arguments normally passed in a message.
         */
        __block NSUInteger index = 2;
        NSArray<NSObject *> *arguments = call.arguments;
        [arguments enumerateObjectsUsingBlock:^(NSObject * _Nonnull arg, NSUInteger idx, BOOL * _Nonnull stop) {
            [invocation setArgument:&(arg) atIndex:index];
            index++;
        }];
        
        [invocation invoke];
        
        NSMethodSignature *signature = [invocation methodSignature];
        const char *type = [signature methodReturnType];
        
        if (strcmp(type, "v") != 0) {
            void *returnValue;
            [invocation getReturnValue:&returnValue];
            NSObject *resultSet = (__bridge NSObject *)returnValue;
            result(resultSet);
        }
    }
    
    if (!isImplemented) {
        result(FlutterMethodNotImplemented);
    }
}

- (void)startWithAppKey:(NSString *)appKey clientUserID:(NSString *)clientUserID configurations:(NSDictionary *)configurations {
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    
    NSString *environment = [configurations objectForKey:@"environment"];
    DAPIEnvironment dapiEnvironment = DAPIEnvironmentSandbox;
    if (environment != (id)[NSNull null] &&
        environment != nil &&
        [environment isEqualToString: @"production"]) {
        dapiEnvironment = DAPIEnvironmentProduction;
    }
    [DPC.sharedInstance startWithAppKey:appKey
                            environment:dapiEnvironment
                           clientUserID:clientUserID
                             completion:^(Dapi * _Nullable dapi , NSError * _Nullable error ) {
        if (dapi) {
            [DPCCrossSDK setNativeConfigurations:configurations];
            self.isDapiStarted = YES;
            self.cachedAccounts = [[NSMutableDictionary<NSString *, DAPIBankAccount *> alloc ] init];
            flutterResult(nil);
        } else {
            flutterResult([self flutterErrorFromNSError:error]);
        }
    }];
    
    
}

- (void)setConfigurations:(NSDictionary<NSString *, id> *)configs {
    [DPCCrossSDK setNativeConfigurations:configs];
}

- (void)configurations {
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    NSDictionary<NSString *, id> *jsConfigurations = [DPCCrossSDK jsConfigurations];
    flutterResult(jsConfigurations);
}


- (void)isStarted {
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    flutterResult([NSNumber numberWithBool:self.isDapiStarted]);
}

- (void)presentConnect:(NSString * _Nullable )bankID {
    dispatch_async(dispatch_get_main_queue(), ^{
        if (bankID != (id)[NSNull null] && bankID != nil ) {
            self.connectVC = [[DAPIConnectVC alloc] initWithBankID:bankID];
        } else {
            self.connectVC = [[DAPIConnectVC alloc] initWithBankID:nil];
        }
        self.connectVC.delegate = self;
        [DPCCrossSDK.topViewController presentViewController:self.connectVC
                                                    animated:true
                                                  completion:nil];
    });
}

- (void)presentAccountSelection:(NSString *)userID {
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    DAPIBankConnection *bankConnection = [DPCCrossSDK  bankConnectionForUserID:userID];
    self.bankAccountsVC = [[DAPIBankAccountsVC alloc] initWithBankConnection:bankConnection];
    
    [self.bankAccountsVC setAccountDidSelect:^(DAPIBankAccount * _Nonnull account) {
        flutterResult(account.id);
    }];
    __weak DapiPlugin *weakSelf = self;
    [self.bankAccountsVC setAccountSelectionFailed:^(NSError * _Nullable error) {
        flutterResult([weakSelf flutterErrorFromNSError:error]);
    }];
    
    [DPCCrossSDK.topViewController presentViewController:self.bankAccountsVC
                                                animated:true
                                              completion:nil];
}

- (void)dismissConnect {
    dispatch_async(dispatch_get_main_queue(), ^{
        if (self.connectVC) {
            [self.connectVC dismissViewControllerAnimated:true completion:nil];
        }
    });
}

- (void)dismissTransfer {
    dispatch_async(dispatch_get_main_queue(), ^{
        if (self.autoFlow) {
            [[self autoFlow] dismissViewControllerAnimated:YES completion:nil];
        }
    });
}

- (void)dismissMFA {
    dispatch_async(dispatch_get_main_queue(), ^{
        if ([[DPCCrossSDK topViewController].navigationController isKindOfClass:[MFANavigationController class]]) {
            MFANavigationController *nav = ((MFANavigationController *) [DPCCrossSDK topViewController].navigationController);
            [nav dismissSelf];
        }
    });
}

- (void)setClientUserID:(NSString *)clientUserID {
    Dapi.shared.clientUserID = clientUserID;
}

- (void)clientUserID {
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    flutterResult(Dapi.shared.clientUserID);
}

- (NSArray<NSDictionary<NSString *, id> *> *) getConnections {
    NSMutableArray *arrayObjects = [NSMutableArray array];
    [DPC.sharedInstance.getConnections enumerateObjectsUsingBlock:^(DAPIBankConnection * _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        [arrayObjects addObject: [obj dictionaryRepresentation]];
    }];
    return arrayObjects;
}

- (void)createConnection:(NSString *)jsonConnectionParameters {
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    NSData *jsonData = [jsonConnectionParameters dataUsingEncoding:NSUTF8StringEncoding];
    NSError *err;
    NSDictionary<NSString *, id> *connectionParams = [NSJSONSerialization JSONObjectWithData:jsonData options:0 error:&err];
    
    if (err) {
        flutterResult([self flutterErrorFromNSError:err]);
    }
    
    [DPC.sharedInstance createBankConnectionWithParameters:connectionParams
                                                completion:^(NSDictionary<NSString *,id> * _Nullable connection, NSError * _Nullable error) {
        if (error != (id)[NSNull null] &&
            error != nil ) {
            flutterResult([self flutterErrorFromNSError:error]);
        } else {
            flutterResult(connection);
        }
    }];
    
}

- (void)getParameters:(NSString *)userID {
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    DAPIBankConnection *bankConnection = [DPCCrossSDK  bankConnectionForUserID:userID];
    NSDictionary<NSString *, id> *connectionParams = bankConnection.getParameters;
    NSError *error;
    NSData *jsonData = [NSJSONSerialization dataWithJSONObject:connectionParams
                                                       options:NSJSONWritingPrettyPrinted
                                                         error:&error];
    
    if (!jsonData) {
        flutterResult([self flutterErrorFromNSError:error]);
    } else {
        NSString *json = [[NSString alloc] initWithData:jsonData encoding:NSUTF8StringEncoding];
        flutterResult(json);
    }
}

- (void)getIdentity:(NSString *)userID {
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    DAPIBankConnection *bankConnection = [DPCCrossSDK  bankConnectionForUserID:userID];
    [DPC.sharedInstance identityWithBankConnection:bankConnection
                                        completion:^(NSDictionary<NSString *,id> * _Nullable data ,
                                                     NSError * _Nullable error ) {
        [self respondForDictionaryRepresentableObject:data error:error result:flutterResult];
    }];
}

- (void)getAccounts:(NSString *)userID {
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    DAPIBankConnection *bankConnection = [DPCCrossSDK  bankConnectionForUserID:userID];
    [DPC.sharedInstance bankAccountsWithBankConnection:bankConnection
                                            completion:^(NSDictionary<NSString *,id> * _Nullable data ,
                                                            NSError * _Nullable error ) {
        if (data != nil ) {
            NSArray *accounts = [data valueForKey:@"accounts"];
            for (NSDictionary<NSString *,id> *account in accounts)
            {
                NSString *senderBankAccountId = [account valueForKey:@"id"];
                NSDictionary<NSString *,id> *balance = [account valueForKey:@"balance"];
                NSNumber *amount = [balance valueForKey:@"amount"];
                NSDictionary<NSString *,id> *currency = [account valueForKey:@"currency"];
                NSString *code = [currency valueForKey:@"code"];
                
                if (senderBankAccountId && balance ) {
                    DAPIBankAccount *account = [[DAPIBankAccount alloc ] init];
                    DAPIBalance *balance = [[DAPIBalance alloc ] init];
                    [balance setAmount:amount.doubleValue];
                    DAPIPair *currency = [[DAPIPair alloc ] init];
                    [currency setCode:code];
                    [account setBalance:balance];
                    [account setCurrency:currency];
                    [account setId:senderBankAccountId];
                    [self.cachedAccounts setObject:account forKey:senderBankAccountId];
                }
                
            }
        }
        [self respondForDictionaryRepresentableObject:data error:error result:flutterResult];
    }];
}

- (void)getCards:(NSString *)userID {
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    DAPIBankConnection *bankConnection = [DPCCrossSDK  bankConnectionForUserID:userID];
    [DPC.sharedInstance bankCardsWithBankConnection:bankConnection
                                         completion:^(NSDictionary<NSString *,id> * _Nullable data ,
                                                      NSError * _Nullable error ) {
        [self respondForDictionaryRepresentableObject:data error:error result:flutterResult];
    }];
}

- (void)getTransactionsForAccount:(NSString *)userID accountID:(NSString *)accountID startDate:(NSNumber *)startDate endDate:(NSNumber *)endDate type:(NSString *)type {
    NSDate *fromDate = [NSDate dateWithTimeIntervalSince1970:startDate.doubleValue/1000];
    NSDate *toDate = [NSDate dateWithTimeIntervalSince1970:endDate.doubleValue/1000];
    DAPITransactionType enumType = DAPITransactionTypeDefault;
    if([type isEqualToString:@"enriched"]) {
        enumType = DAPITransactionTypeEnriched;
    } else if([type isEqualToString:@"categorized"]) {
        enumType = DAPITransactionTypeCategorized;
    }
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];

    DAPIBankConnection *bankConnection = [DPCCrossSDK  bankConnectionForUserID:userID];
    [DPC.sharedInstance accountTransactionsWithBankConnection:bankConnection
                                                bankAccountID:accountID
                                                         type:enumType
                                                         from:fromDate
                                                           to:toDate
                                                   completion:^(NSDictionary<NSString *,id> * _Nullable data ,
                                                                NSError * _Nullable error ) {
        [self respondForDictionaryRepresentableObject:data error:error result:flutterResult];
    }];
}

- (void)getTransactionsForCard:(NSString *)userID cardID:(NSString *)cardID startDate:(NSNumber *)startDate endDate:(NSNumber *)endDate type:(NSString *)type {
    NSDate *fromDate = [NSDate dateWithTimeIntervalSince1970:startDate.doubleValue/1000];
    NSDate *toDate = [NSDate dateWithTimeIntervalSince1970:endDate.doubleValue/1000];

    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    DAPITransactionType enumType = DAPITransactionTypeDefault;
    if([type isEqualToString:@"enriched"]) {
        enumType = DAPITransactionTypeEnriched;
    } else if([type isEqualToString:@"categorized"]) {
        enumType = DAPITransactionTypeCategorized;
    }

    DAPIBankConnection *bankConnection = [DPCCrossSDK  bankConnectionForUserID:userID];
    [DPC.sharedInstance cardTransactionsWithBankConnection:bankConnection
                                                    cardID:cardID
                                                      type:enumType
                                                      from:fromDate
                                                        to:toDate
                                                completion:^(NSDictionary<NSString *,id> * _Nullable data ,
                                                             NSError * _Nullable error ) {
        [self respondForDictionaryRepresentableObject:data error:error result:flutterResult];
    }];
}

- (void)getAccountsMetadata:(NSString *)userID {
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    DAPIBankConnection *bankConnection = [DPCCrossSDK  bankConnectionForUserID:userID];
    [DPC.sharedInstance accountMetadataWithBankConnection:bankConnection
                                               completion:^(NSDictionary<NSString *,id> * _Nullable data ,
                                                            NSError * _Nullable error ) {
        [self respondForDictionaryRepresentableObject:data error:error result:flutterResult];
    }];
}

- (void)delete:(NSString *)userID {
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    DAPIBankConnection *bankConnection = [DPCCrossSDK  bankConnectionForUserID:userID];
    [DPC.sharedInstance delinkWithBankConnection:bankConnection
                                      completion:^(NSDictionary<NSString *,id> * _Nullable data ,
                                                   NSError * _Nullable error ) {
        [self respondForDictionaryRepresentableObject:data error:error result:flutterResult];
    }];
}

- (void)getBeneficiaries:(NSString *)userID {
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    DAPIBankConnection *bankConnection = [DPCCrossSDK  bankConnectionForUserID:userID];
    [DPC.sharedInstance bankBeneficiariesWithBankConnection:bankConnection
                                                 completion:^(NSDictionary<NSString *,id> * _Nullable data ,
                                                              NSError * _Nullable error ) {
        [self respondForDictionaryRepresentableObject:data error:error result:flutterResult];
    }];
}

- (void)getWireBeneficiaries:(NSString *)userID {
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    DAPIBankConnection *bankConnection = [DPCCrossSDK  bankConnectionForUserID:userID];
    [DPC.sharedInstance bankWireBeneficiariesWithBankConnection:bankConnection
                                                     completion:^(NSDictionary<NSString *,id> * _Nullable data ,
                                                                  NSError * _Nullable error ) {
        [self respondForDictionaryRepresentableObject:data error:error result:flutterResult];
    }];
}

- (void)createBeneficiary:(NSString *)userID beneficiary:(NSDictionary<NSString *, id> *)beneficiary {
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    DAPIBankConnection *bankConnection = [DPCCrossSDK  bankConnectionForUserID:userID];
    DAPIBeneficiary *nativeBeneficiary = [DPCCrossSDK nativeBeneficiaryInfoFromDictionary:beneficiary];
    [DPC.sharedInstance createBeneficiaryWithBankConnection:bankConnection
                                         beneficiaryDetails:nativeBeneficiary
                                                 completion:^(NSDictionary<NSString *,id> * _Nullable data ,
                                                              NSError * _Nullable error ) {
        [self respondForDictionaryRepresentableObject:data error:error result:flutterResult];
    }];

}
- (void)createWireBeneficiary:(NSString *)userID beneficiary:(NSDictionary<NSString *, id> *)beneficiary {
    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    DAPIBankConnection *bankConnection = [DPCCrossSDK  bankConnectionForUserID:userID];
    DAPIWireBeneficiary *nativeBeneficiary = [DPCCrossSDK nativeWireBeneficiaryInfoFromDictionary:beneficiary];
    [DPC.sharedInstance createWireBeneficiaryWithBankConnection:bankConnection
                                             beneficiaryDetails:nativeBeneficiary
                                                     completion:^(NSDictionary<NSString *,id> * _Nullable data ,
                                                                  NSError * _Nullable error ) {
        [self respondForDictionaryRepresentableObject:data error:error result:flutterResult];
    }];
}

- (void)createTransfer:(NSString *)userID
             accountID:(NSString *)accountID
           beneficiary:(NSDictionary<NSString *, id> *)beneficiary
                amount:(NSNumber *)amount
                remark:(NSString *)remark {
    
    dispatch_async(dispatch_get_main_queue(), ^{
        NSString *resultKey = NSStringFromSelector(_cmd);
        FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
        NSString *_remark = remark;
        if (_remark == (id)[NSNull null] || _remark == nil ) { _remark = nil; }

        DAPIBankConnection *bankConnection = [DPCCrossSDK bankConnectionForUserID:userID];
        
        DAPIBeneficiary *nativeBeneficiary;
        if(NULL_TO_NIL(beneficiary) == nil) {
            nativeBeneficiary = nil;
        } else {
            nativeBeneficiary = [DPCCrossSDK nativeBeneficiaryInfoFromDictionary:beneficiary];
        }
        
        [self createTransferWithBankConnection:bankConnection
                           senderBankAccountId:accountID
                                 toBeneficiary:nativeBeneficiary
                                        amount:[amount doubleValue]
                                        remark:_remark
                                    completion:^(NSDictionary<NSString *,id> * _Nullable data ,
                                                 NSError * _Nullable error ) {
            if (error) {
                [self respondForDictionaryRepresentableObject:nil error:error result:flutterResult];
            } else {
                [self respondForDictionaryRepresentableObject:
                 @{
                    @"account": [[NSString alloc] initWithFormat:@"%@", data[@"senderAccountID"]],
                    @"amount": [NSNumber numberWithDouble:[amount doubleValue]],
                    @"operationID": [[NSString alloc] initWithFormat:@"%@", data[@"operationID"]],
                    @"remark": [[NSString alloc] initWithFormat:@"%@", data[@"remark"]] ?: [NSNull null],
                    @"reference": [[NSString alloc] initWithFormat:@"%@", data[@"reference"]] ?: [NSNull null]
                } error:nil result:flutterResult];
                
            }
        }];
    });
    
}

- (void)createWireTransfer:(NSString *)userID accountID:(NSString *)accountID beneficiary:(NSDictionary<NSString *, id> *)beneficiary amount:(NSNumber *)amount remark:(NSString *)remark {

    
    dispatch_async(dispatch_get_main_queue(), ^{
        NSString *resultKey = NSStringFromSelector(_cmd);
        FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
        NSString *_remark = remark;
        if (_remark == (id)[NSNull null] || _remark == nil ) { _remark = nil; }

        DAPIBankConnection *bankConnection = [DPCCrossSDK  bankConnectionForUserID:userID];
        DAPIWireBeneficiary *nativeBeneficiary = [DPCCrossSDK nativeWireBeneficiaryInfoFromDictionary:beneficiary];

        [self createWireTransferWithBankConnection:bankConnection
                               senderBankAccountId:accountID
                                     toBeneficiary:nativeBeneficiary
                                            amount:[amount doubleValue]
                                            remark:_remark
                                        completion:^(NSDictionary<NSString *,id> * _Nullable data ,
                                                     NSError * _Nullable error ) {
            if (error) {
                [self respondForDictionaryRepresentableObject:nil error:error result:flutterResult];
            } else {
                [self respondForDictionaryRepresentableObject:
                 @{
                    @"account": [[NSString alloc] initWithFormat:@"%@", data[@"senderAccountID"]],
                    @"amount": [NSNumber numberWithDouble:[amount doubleValue]],
                    @"operationID": [[NSString alloc] initWithFormat:@"%@", data[@"operationID"]],
                    @"remark": [[NSString alloc] initWithFormat:@"%@", data[@"remark"]] ?: [NSNull null],
                    @"reference": [[NSString alloc] initWithFormat:@"%@", data[@"reference"]] ?: [NSNull null]
                } error:nil result:flutterResult];
                
            }
        }];
    });
    
}

- (void)createWireTransferToExistingBeneficiary:(NSString *)userID accountID:(NSString *)accountID receiverID:(NSString *)receiverID amount:(NSNumber *)amount remark:(NSString *)remark {
    dispatch_async(dispatch_get_main_queue(), ^{
        NSString *resultKey = NSStringFromSelector(_cmd);
        NSString *_remark = remark;
        if (_remark == (id)[NSNull null] || _remark == nil ) { _remark = nil; }

        FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
        DAPIBankConnection *bankConnection = [DPCCrossSDK  bankConnectionForUserID:userID];
        [DPC.sharedInstance createWireTransferToExistBeneficiaryWithBankConnection:bankConnection
                                                               senderBankAccountID:accountID
                                                             receiverBeneficiaryID:receiverID
                                                                            amount:[amount doubleValue]
                                                                            remark:_remark
                                                                        completion:^(NSDictionary<NSString *,id> * _Nullable data ,
                                                                                     NSError * _Nullable error ) {
            if (error) {
                [self respondForDictionaryRepresentableObject:nil error:error result:flutterResult];
            } else {
                [self respondForDictionaryRepresentableObject:
                 @{
                    @"account": [[NSString alloc] initWithFormat:@"%@", data[@"senderAccountID"]],
                    @"amount": [NSNumber numberWithDouble:[amount doubleValue]],
                    @"operationID": [[NSString alloc] initWithFormat:@"%@", data[@"operationID"]],
                    @"remark": [[NSString alloc] initWithFormat:@"%@", data[@"remark"]] ?: [NSNull null],
                    @"reference": [[NSString alloc] initWithFormat:@"%@", data[@"reference"]] ?: [NSNull null]
                } error:nil result:flutterResult];
                
            }
        }];
    });
}

- (void)createTransferToExistingBeneficiary:(NSString *)userID accountID:(NSString *)accountID receiverID:(NSString *)receiverID amount:(NSNumber *)amount remark:(NSString *)remark {

    NSString *resultKey = NSStringFromSelector(_cmd);
    FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
    

    dispatch_async(dispatch_get_main_queue(), ^{
        NSString *_remark = remark;
        if (_remark == (id)[NSNull null] || _remark == nil ) { _remark = nil; }

        DAPIBankConnection *bankConnection = [DPCCrossSDK  bankConnectionForUserID:userID];
        [DPC.sharedInstance createTransferToExistBeneficiaryWithBankConnection:bankConnection
                                                           senderBankAccountID:accountID
                                                         receiverBeneficiaryID:receiverID
                                                                        amount:[amount doubleValue]
                                                                        remark:_remark
                                                                    completion:^(NSDictionary<NSString *,id> * _Nullable data ,
                                                                                 NSError * _Nullable error) {
            if (error) {
                [self respondForDictionaryRepresentableObject:nil error:error result:flutterResult];
            } else {
                [self respondForDictionaryRepresentableObject:
                 @{
                    @"account": [[NSString alloc] initWithFormat:@"%@", data[@"senderAccountID"]],
                    @"amount": [NSNumber numberWithDouble:[amount doubleValue]],
                    @"operationID": [[NSString alloc] initWithFormat:@"%@", data[@"operationID"]],
                    @"remark": [[NSString alloc] initWithFormat:@"%@", data[@"remark"]] ?: [NSNull null],
                    @"reference": [[NSString alloc] initWithFormat:@"%@", data[@"reference"]] ?: [NSNull null]
                } error:nil result:flutterResult];
                
            }
        }];
    });
    
}

- (void)createACHPullTransferWithBankConnection:(DAPIBankConnection * _Nonnull)bankConnection
                            senderBankAccountId:(NSString * _Nonnull)senderBankAccountId
                                         amount:(double)amount
                                    description:(NSString *__nullable)description
                                     completion:(void (^ _Nonnull)(NSDictionary<NSString *,id> * _Nullable data ,
                                                                   NSError * _Nullable error))completion {
    
    BOOL hasAmount = amount != 0 && amount != NSNotFound;
    BOOL hasAccount = senderBankAccountId != nil;
    if (hasAccount && hasAmount) {
        
        [DPC.sharedInstance createACHPullTransferWithBankConnection:bankConnection
                                                senderBankAccountId:senderBankAccountId
                                                             amount:amount
                                                        description:description
                                                         completion:completion];
        
        
    } else {
        DAPIBankAccount *account;
        if ((senderBankAccountId != (id)[NSNull null]) && (senderBankAccountId != nil)) {
            account = [self.cachedAccounts valueForKey:senderBankAccountId ];
        }
        self.autoFlow = [[DAPIAutoFlowVC alloc] initWithBankConnectionForACHPullTransfer:bankConnection
                                                                                 account:account amount:amount description:description];
        [self.autoFlow setTransferDidFail:^(NSError * _Nonnull error) {
            completion(nil,error);
        }];
        
        [self.autoFlow setTransferDidSucceed:^(DAPITransferResult * _Nonnull data) {
            completion(data.dictionaryRepresentation,nil);
            
        }];
        
        [DPCCrossSDK.topViewController presentViewController: [self autoFlow]
                                                    animated:YES
                                                  completion:nil];
        
    }
    
}


- (void)createACHPullTransfer:(NSString *)userID
                  description:(NSString *)description
                  accountID:(NSString *)accountID
                  amount:(NSNumber *)amount {
    
    dispatch_async(dispatch_get_main_queue(), ^{
        NSString *resultKey = NSStringFromSelector(_cmd);
        FlutterResult flutterResult = [self.selectorResult objectForKey:resultKey];
        NSString *_description = description;
        if (_description == (id)[NSNull null] || _description == nil ) { _description = nil; }
        
        DAPIBankConnection *bankConnection = [DPCCrossSDK bankConnectionForUserID:userID];
        
        [self createACHPullTransferWithBankConnection:bankConnection
                                  senderBankAccountId:accountID
                                               amount:[amount doubleValue]
                                          description:_description
                                           completion:^(NSDictionary<NSString *,id> * _Nullable data ,
                                                        NSError * _Nullable error ) {
            if (error) {
                [self respondForDictionaryRepresentableObject:nil error:error result:flutterResult];
            } else {
                [self respondForDictionaryRepresentableObject:
                 @{
                    @"account": [[NSString alloc] initWithFormat:@"%@", data[@"senderAccountID"]],
                    @"amount": [NSNumber numberWithDouble:[amount doubleValue]],
                    @"operationID": [[NSString alloc] initWithFormat:@"%@", data[@"operationID"]],
                    @"remark": [[NSString alloc] initWithFormat:@"%@", data[@"remark"]] ?: [NSNull null],
                    @"reference": [[NSString alloc] initWithFormat:@"%@", data[@"reference"]] ?: [NSNull null]
                } error:nil result:flutterResult];
                
            }
        }];
    });
    
}


// MARK: - Helper Methods

- (void)createTransferWithBankConnection:(DAPIBankConnection * _Nonnull)bankConnection
                     senderBankAccountId:(NSString * _Nonnull)senderBankAccountId
                           toBeneficiary:(DAPIBeneficiary *__nullable)beneficiary
                                  amount:(double)amount
                                  remark:(NSString *__nullable)remark
                              completion:(void (^ _Nonnull)(NSDictionary<NSString *,id> * _Nullable data ,
                                                            NSError * _Nullable error))completion {
    NSString *_remark = remark;
    if (_remark == (id)[NSNull null] || _remark == nil ) { _remark = nil; }

    BOOL hasAmount = amount != 0 && amount != NSNotFound;
    BOOL hasAccount = (senderBankAccountId != (id)[NSNull null]) && (senderBankAccountId != nil);
    if (hasAccount && hasAmount) {
        
        [DPC.sharedInstance createTransferWithBankConnection:bankConnection
                                         senderBankAccountId:senderBankAccountId
                                         receiverBeneficiary:beneficiary
                                                      amount:amount
                                                      remark:_remark
                                                  completion:completion];
        
    } else {
        DAPIBankAccount *account;
        if ((senderBankAccountId != (id)[NSNull null]) && (senderBankAccountId != nil) ) {
            account = [self.cachedAccounts valueForKey:senderBankAccountId ];
        }
        self.autoFlow = [[DAPIAutoFlowVC alloc] initWithBankConnection:bankConnection
                                                               account:account
                                                       wireBeneficiary: nil
                                                           beneficiary:beneficiary
                                                          transferType:DAPITransferTypeNormal
                                                                amount:amount
                                                                remark:_remark];
        [self.autoFlow setTransferDidFail:^(NSError * _Nonnull error) {
            completion(nil,error);
        }];
        
        [self.autoFlow setTransferDidSucceed:^(DAPITransferResult * _Nonnull data) {
            completion(data.dictionaryRepresentation,nil);
            
        }];
        [self.autoFlow setAutoFlowDelegate:self];
        
        [DPCCrossSDK.topViewController presentViewController: [self autoFlow]
                                                    animated:YES
                                                  completion:nil];
        
    }
    
}

- (void)createWireTransferWithBankConnection:(DAPIBankConnection * _Nonnull)bankConnection
                         senderBankAccountId:(NSString * _Nonnull) senderBankAccountId
                               toBeneficiary:(DAPIWireBeneficiary *__nullable)beneficiary
                                      amount:(double) amount
                                      remark:(NSString *__nullable)remark
                                  completion:(void (^ _Nonnull)(NSDictionary<NSString *,id> * _Nullable data ,
                                                                NSError * _Nullable error))completion {
    NSString *_remark = remark;
    if (_remark == (id)[NSNull null] || _remark == nil ) { _remark = nil; }

    BOOL hasAmount = amount != 0 && amount != NSNotFound;
    BOOL hasAccount = (senderBankAccountId != (id)[NSNull null]) && (senderBankAccountId != nil);
    if (hasAccount && hasAmount) {
        
        [DPC.sharedInstance createWireTransferWithBankConnection:bankConnection
                                             senderBankAccountID:senderBankAccountId
                                             receiverBeneficiary:beneficiary
                                                          amount:amount
                                                          remark:_remark
                                                      completion:completion];
        
    } else {
        DAPIBankAccount *account;
        if ((senderBankAccountId != (id)[NSNull null]) && (senderBankAccountId != nil) ) {
            account = [self.cachedAccounts valueForKey:senderBankAccountId ];
        }
        self.autoFlow = [[DAPIAutoFlowVC alloc] initWithBankConnection:bankConnection
                                                               account:account
                                                       wireBeneficiary: beneficiary
                                                           beneficiary:nil
                                                          transferType:DAPITransferTypeWire
                                                                amount:amount
                                                                remark:_remark];
        [self.autoFlow setTransferDidFail:^(NSError * _Nonnull error) {
            completion(nil,error);
        }];
        
        [self.autoFlow setTransferDidSucceed:^(DAPITransferResult * _Nonnull data) {
            completion(data.dictionaryRepresentation,nil);
        }];
        [self.autoFlow setAutoFlowDelegate:self];

        [DPCCrossSDK.topViewController presentViewController: [self autoFlow]
                                                    animated:YES
                                                  completion:nil];
        
    }
    
}

- (void)respondForDictionaryRepresentableObject:(NSObject *)object error:(NSError *)error result:(FlutterResult)result {
    
    if (error) {
        result([self flutterErrorFromNSError:error]);
    } else if (object != (id)[NSNull null] && object != nil ) {
        result(object);
    } else {
        NSError *castingError = [NSError errorWithDomain:@"com.dapi.dapiconnect.flutter" code:1012 userInfo:@{NSLocalizedDescriptionKey: @"Couldn't construct JSON representation of native array objects"}];
        result([self flutterErrorFromNSError:castingError]);
    }
}

- (FlutterError *)flutterErrorFromNSError:(NSError *)error {
    NSDictionary *jsonError = @{
        @"message": [error dapiErrorMessage] ?: [NSNull null],
        @"type": [error dapiErrorType] ?: [NSNull null],
        @"statusCode": error.dapiStatusCode  ?: [NSNull null],
        @"operationID": error.dapiOperationID ?: [NSNull null],
    };

    NSDictionary *coolDownPeriod = [error.dictionaryRepresentation objectForKey:@"coolDownPeriod"];
    if ( NULL_TO_NIL(coolDownPeriod) != nil) {
        NSNumber *value = [coolDownPeriod objectForKey:@"value"] ?: [NSNull null];
        NSString *unit = [coolDownPeriod objectForKey:@"unit"]?: [NSNull null];
        jsonError = @{
            @"message": [error dapiErrorMessage] ?: [NSNull null],
            @"type": [error dapiErrorType] ?: [NSNull null],
            @"statusCode": error.dapiStatusCode  ?: [NSNull null],
            @"operationID": error.dapiOperationID ?: [NSNull null],
            @"value": value,
            @"unit": unit

        };
    }

    NSError *convertingError;
    NSData *jsonData = [NSJSONSerialization dataWithJSONObject:jsonError
                                                       options:NSJSONWritingPrettyPrinted
                                                         error:&convertingError];
    
    if (convertingError) {
        FlutterError *flutterError = [FlutterError errorWithCode:[NSString stringWithFormat:@"%@", [@(error.code) stringValue]] message:[NSString stringWithFormat:@"{\"message\": \"%@\"}", error.dapiErrorMessage] details:nil];
        return flutterError;
    } else {
        FlutterError *flutterError = [FlutterError errorWithCode:[NSString stringWithFormat:@"%@", [@(error.code) stringValue]] message:[[NSString alloc] initWithData:jsonData encoding:NSUTF8StringEncoding] details:nil];
        return flutterError;
    }
}


- (void)autoFlow:(DAPIAutoFlowVC * _Nonnull)autoFlow
transferDidSuccessFrom:(DAPIBankAccount * _Nonnull)senderAccount
  transferResult:(DAPITransferResult * _Nonnull)transferResult { }

- (void)autoFlow:(DAPIAutoFlowVC * _Nonnull)autoFlow willTransferAmount:(float)amount fromAccount:(DAPIBankAccount * _Nonnull)senderAccount {
    NSDictionary * dic = [NSDictionary dictionaryWithObjectsAndKeys:
                          [NSNumber numberWithDouble:amount],@"amount",
                          NULL_TO_NIL(senderAccount.id),@"account", nil];
    [self.transferUiWillSendStream send: dic];

}

- (void)autoFlowUserDidCanceled:(DAPIAutoFlowVC * _Nonnull)autoFlow { }

@end

@implementation SuccessfulConnectionStream {
    FlutterEventSink _eventSink;
}

- (FlutterError *_Nullable)onListenWithArguments:(id _Nullable)arguments eventSink:(FlutterEventSink)events {
    _eventSink = events;
    return nil;
}

- (void)send:(NSDictionary *)data {
    if (_eventSink) {
        _eventSink(data);
    }
}

- (FlutterError *_Nullable)onCancelWithArguments:(id _Nullable)arguments {
    _eventSink = nil;
    return nil;
}

@end

@implementation FailureConnectionStream {
    FlutterEventSink _eventSink;
}

- (FlutterError *_Nullable)onListenWithArguments:(id _Nullable)arguments eventSink:(FlutterEventSink)events {
    _eventSink = events;
    return nil;
}

- (void)send:(NSDictionary *)data {
    if (_eventSink) {
        _eventSink(data);
    }
}

- (FlutterError *_Nullable)onCancelWithArguments:(id _Nullable)arguments {
    _eventSink = nil;
    return nil;
}

@end

@implementation BankRequestStream {
    FlutterEventSink _eventSink;
}

- (FlutterError *_Nullable)onListenWithArguments:(id _Nullable)arguments eventSink:(FlutterEventSink)events {
    _eventSink = events;
    return nil;
}

- (void)send:(NSDictionary *)data {
    if (_eventSink) {
        _eventSink(data);
    }
}

- (FlutterError *_Nullable)onCancelWithArguments:(id _Nullable)arguments {
    _eventSink = nil;
    return nil;
}

@end

@implementation ConnectDismissedStream {
    FlutterEventSink _eventSink;
}

- (FlutterError *_Nullable)onListenWithArguments:(id _Nullable)arguments eventSink:(FlutterEventSink)events {
    _eventSink = events;
    return nil;
}

- (void)send:(NSDictionary *)data {
    if (_eventSink) {
        _eventSink(data);
    }
}

- (FlutterError *_Nullable)onCancelWithArguments:(id _Nullable)arguments {
    _eventSink = nil;
    return nil;
}

@end

@implementation TransferUiDismissedStream {
    FlutterEventSink _eventSink;
}

- (FlutterError *_Nullable)onListenWithArguments:(id _Nullable)arguments eventSink:(FlutterEventSink)events {
    _eventSink = events;
    return nil;
}

- (void)send:(NSDictionary *)data {
    if (_eventSink) {
        _eventSink(data);
    }
}

- (FlutterError *_Nullable)onCancelWithArguments:(id _Nullable)arguments {
    _eventSink = nil;
    return nil;
}

@end

@implementation TransferUiWillSendStream {
    FlutterEventSink _eventSink;
}

- (FlutterError *_Nullable)onListenWithArguments:(id _Nullable)arguments eventSink:(FlutterEventSink)events {
    _eventSink = events;
    return nil;
}

- (void)send:(NSDictionary *)data {
    if (_eventSink) {
        _eventSink(data);
    }
}

- (FlutterError *_Nullable)onCancelWithArguments:(id _Nullable)arguments {
    _eventSink = nil;
    return nil;
}

@end
